﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Net;
using System.Security;
using System.Threading.Tasks;
using WolframAlpha.WrapperCore;

namespace SimpleCrawler
{
    public class Crawler
    {
        #region Fields

        /// <summary>
        /// List of querys
        /// </summary>
        public IList<string> Querys
        {
            get;
            set;
        }

        /// <summary>
        /// Search engine
        /// </summary>
        private readonly WolframAlphaEngine _engine;

        /// <summary>
        /// Results of search
        /// </summary>
        public IList<WolframAlphaQueryResult> Results
        {
            get;
            set;
        }

        /// <summary>
        /// Path to folder, for results write
        /// </summary>
        private static string _path;

        /// <summary>
        /// Formats of output data
        /// </summary>
        private readonly List<string> _formats;

        /// <summary>
        /// ApplicationID
        /// </summary>
        public string AppID
        {
            get;
            set;
        }
        #endregion

        #region Constructors
        #region Private constructor
        /// <summary>
        /// Dump protection
        /// </summary>
        private Crawler()
        {

        }
        #endregion

        #region Public constructor(IList<string>, string, IList<string>)
        /// <summary>
        /// Constructor (IList<string/>, string, IList<string/>)
        /// </summary>
        /// <param name="querys">list or array or something else contains query strings</param>
        /// <param name="appid">Application Id</param>
        /// <param name="formats"></param>
        public Crawler(IList<string> querys, string appid, IList<string> formats = null)
        {
            #region Conditions
            //I'm not paranoid, I just check everything
            if (querys == null || querys.Count == 0)
            {
                return;
            }
            if (string.IsNullOrEmpty(appid))
            {
                throw new ArgumentNullException(Properties.Resources.ApplicationID_is_null);
            }
            #endregion

            #region Allocation/Init
            if (formats != null && formats.Count != 0)
            {
                _formats = formats.ToList();
            }
            else
            {
                _formats = new List<string>();
            }
            AppID = appid;
            Querys = querys;
            _engine = new WolframAlphaEngine(appid);
            Results = new List<WolframAlphaQueryResult>();
            #endregion
        }
        #endregion
        #endregion

        #region Methods

        #region Public Methods

        #region SearchData
        /// <summary>
        /// SearchData
        /// </summary>
        public void SearchData(bool isappend = false)
        {
            #region Conditions
            //I'm not paranoid, I just check everything
            if (Querys == null || Querys.Count == 0)
            {
                return;
            }
            if (string.IsNullOrEmpty(AppID))
            {
                throw new NullReferenceException("ApplicationID is null");
            }
            #endregion

            #region Search
            var tasklist = new Task[Querys.Count];
            int i = 0;
            foreach (var querystr in Querys)
            {
                try
                {
                    var query = new WolframAlphaQuery
                    {
                        APIKey = _engine.APIKey,
                        Formats = _formats,
                        Query = querystr
                    };
                    int count = i;
                    tasklist[i] = Task.Factory.StartNew(() => GetWolframAlphaResultAsync(query, count, isappend));
                    i++;
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message);
                }
            }
            Task.WaitAll(tasklist);
            #endregion
        }
        #endregion

        #region WriteLogAll
        /// <summary>
        /// Write all result in one time
        /// </summary>
        public void WriteLogAll()
        {
            #region Conditions
            if (Results == null || Results.Count == 0)
            {
                return;
            }
            #endregion

            int count = 0;
            foreach (var wolframAlphaQueryResult in Results)
            {
                WriteOneResult(wolframAlphaQueryResult, count);
                count++;
            }
        }
        #endregion
        #endregion

        #region Private Methods

        #region GetWolframAlphaResultAsync
        /// <summary>
        /// GetWolframAlphaResultAsync
        /// </summary>
        /// <param name="query">Query</param>
        /// <param name="count">Number</param>
        /// <param name="isappend">Is append to file</param>
        /// <returns></returns>
        private void GetWolframAlphaResultAsync(WolframAlphaQuery query, int count, bool isappend)
        {
            var newengine = new WolframAlphaEngine(_engine.APIKey);
            var result = newengine.GetWolframAlphaQueryResult(newengine.GetStringRequest(query));
            if (isappend)
            {
                WriteOneResult(result, count);
            }
            else
            {
                Results.Add(result);
            }
            Console.WriteLine(query.Query);
        }
        #endregion

        #region WriteOneResult
        /// <summary>
        /// Write Result of one search query
        /// </summary>
        /// <param name="wolframAlphaQueryResult">Result of query</param>
        /// <param name="count">Number of result</param>
        private static void WriteOneResult(WolframAlphaQueryResult wolframAlphaQueryResult, int count = 0)
        {
            #region Conditions
            if (wolframAlphaQueryResult == null)
            {
                return;
            }
            #endregion

            try
            {
                string filepath = Properties.Settings.Default.PathToOutputDirectory + "\\" + wolframAlphaQueryResult.Query.Replace(" ", "")+".txt";
                Directory.CreateDirectory(filepath);
                _path = filepath;
                foreach (var item in wolframAlphaQueryResult.Pods)
                {
                    WriteResultOutputLog("");
                    WriteResultOutputLog("Sub pods found: " + item.NumberOfSubPods);
                    WriteResultOutputLog("Title: \"" + item.Title + "\"");
                    WriteResultOutputLog("Position: " + item.Position);
                    int subPodNumber = 1;

                    foreach (var subItem in item.SubPods)
                    {
                        try
                        {
                            var webRequest = (HttpWebRequest)(WebRequest.Create(subItem.PodImage.Location.ToString()));
                            webRequest.KeepAlive = true;
                            var b = new Bitmap(webRequest.GetResponse().GetResponseStream());
                            b.Save(_path + "\\" + count + ".png");
                            count++;
                        }
                        catch (NullReferenceException)
                        {
                            WriteProblemResult(Properties.Resources.bitmap_from_url);
                        }
                        catch (Exception)
                        {
                            WriteProblemResult(Properties.Resources.SomeProblem);
                        }
                        WriteResultOutputLog("");
                        WriteResultOutputLog("Sub pod " + subPodNumber);
                        WriteResultOutputLog("Title: \"" + subItem.Title + "\"");
                        WriteResultOutputLog("Pod text: \"" + subItem.PodText + "\"");
                        WriteResultOutputLog("Pod image title: \"" + subItem.PodImage.Title + "\"");
                        WriteResultOutputLog("Pod image width: " + subItem.PodImage.Width);
                        WriteResultOutputLog("Pod image height: " + subItem.PodImage.Height);
                        WriteResultOutputLog("Pod image location: \"" + subItem.PodImage.Location + "\"");
                        WriteResultOutputLog("Pod image description text: \"" + subItem.PodImage.HoverText + "\"");
                        subPodNumber += 1;
                    }
                }
            }
            catch (UnauthorizedAccessException ex)
            {
                WriteProblemResult(ex.Message);
            }
            catch (ArgumentException ex)
            {
                WriteProblemResult(ex.Message);
            }
            catch (DirectoryNotFoundException ex)
            {
                WriteProblemResult(ex.Message);
            }
            catch (PathTooLongException ex)
            {
                WriteProblemResult(ex.Message);
            }
            catch (IOException ex)
            {
                WriteProblemResult(ex.Message);
            }
            catch (SecurityException ex)
            {
                WriteProblemResult(ex.Message);
            }
            catch (Exception ex)
            {
                WriteProblemResult(ex.Message);
            }
        }
        #endregion

        #region WriteResultOutputLog
        /// <summary>
        /// Write Result in Output Log
        /// </summary>
        /// <param name="data">Data for writing</param>
        private static void WriteResultOutputLog(string data)
        {
            #region Conditions
            if (string.IsNullOrEmpty(data))
            {
                return;
            }
            #endregion

            var writer = new StreamWriter(_path + "\\WolframAlphawrapperlog.log", true);
            ConsoleLoggin(data);

            writer.WriteLine(data);
            writer.Close();
            writer.Dispose();
        }
        #endregion

        #region ConsoleLoggin
        /// <summary>
        /// ConsoleLoggin
        /// </summary>
        /// <param name="data"></param>
        private static void ConsoleLoggin(string data)
        {
            Console.ForegroundColor = ConsoleColor.Green;
            Console.WriteLine(data);
        }
        #endregion

        #region WriteProblemResult
        /// <summary>
        /// WriteProblemResult in Console
        /// </summary>
        /// <param name="message">Message to write</param>
        private static void WriteProblemResult(string message)
        {
            Console.ForegroundColor = ConsoleColor.Red;
            Console.WriteLine(message);
            Console.ForegroundColor = ConsoleColor.Green;
        }
        #endregion
        #endregion
        #endregion
    }
}
