﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Net;
using System.IO;
#if MONOTOUCH || MONODROID ||NET
using System.Xml;
#endif

#if NET
using System.Xml.XPath;
#endif

#if NETFX_CORE
using Windows.Data.Xml.Dom;
using System.Linq;
#endif
#if WINDOWS_PHONE
using System.Text;
using System.Xml;
using System.Linq;
using System.Xml.Linq;
#endif
namespace WolframAlpha.WrapperCore
{
    /// <summary>
    /// Class witch request, and parse response
    /// </summary>
    public partial class WolframAlphaEngine
    {
        #region Fiels
        /// <summary>
        /// Aplication key
        /// </summary>
        public string APIKey
        {
            get;
            set;
        }
        /// <summary>
        /// Query Result after request
        /// </summary>
        public WolframAlphaQueryResult QueryResult
        {
            get;
            private set;
        }

        public string QueryResultString
        {
            get;
            private set;
        }

        /// <summary>
        /// Validate Result of response
        /// </summary>
        public WolframAlphaValidationResult ValidationResult
        {
            get;
            private set;
        }
        /// <summary>
        /// Query object
        /// </summary>
        public WolframAlphaQuery Query
        {
            get;
            set;
        }
        #endregion

        #region Constructors

        #region private Constructor
        /// <summary>
        /// private Constructor 
        /// </summary>
        private WolframAlphaEngine()
        {

        }
        #endregion

        #region public Constructor
        /// <summary>
        /// Constructor with aplication key
        /// </summary>
        /// <param name="appkey">Application key</param>
        public WolframAlphaEngine(string appkey)
        {
            if (string.IsNullOrEmpty(appkey))
            {
                throw new ArgumentNullException(@"ApplicationID is null");
            }
            APIKey = appkey;
        }
        #endregion

        #endregion



#if !WINDOWS_PHONE
        /// <summary>
        /// Get Request from service and return result as string
        /// </summary>
        /// <param name="query">Query</param>
        /// <returns>Result as string</returns>   
        public string GetStringRequest(WolframAlphaQuery query)
        {
            if (query == null || string.IsNullOrEmpty(query.APIKey) || string.IsNullOrEmpty(query.Query))
            {
                return string.Empty;
            }

            Query = query;
            string response = string.Empty;
            //Get xml request from the server
            try
            {

                var webRequest = (HttpWebRequest)(WebRequest.Create(query.FullQueryString));
#if MONOTOUCH || NET || MONODROID
                webRequest.KeepAlive = true;
                using (var responseStream = webRequest.GetResponse().GetResponseStream())
                {
                    if (responseStream != null)
                    {
                        response = new StreamReader(responseStream).ReadToEnd();
                        responseStream.Dispose();
                    }
                }
#endif
#if NETFX_CORE
                using (var responseStream = webRequest.GetResponseAsync().Result.GetResponseStream())
                {
                    if (responseStream != null)
                    {
                        response = new StreamReader(responseStream).ReadToEnd();
                    }
                }
#endif
            }
            #region Catch Exceptions
            catch (ProtocolViolationException ex)
            {
                WriteProblemResult(ex.Message);
            }
            catch (NotSupportedException ex)
            {
                WriteProblemResult(ex.Message);
            }
            catch (WebException ex)
            {
                WriteProblemResult(ex.Message);
            }
            catch (InvalidOperationException ex)
            {
                WriteProblemResult(ex.Message);
            }
            catch (Exception ex)
            {
                WriteProblemResult(ex.Message);
            }
            #endregion
            return response;
        }
       
        /// <summary>
        /// GetWolframAlphaQueryResult
        /// </summary>
        /// <param name="response">Response from Request as string</param>
        /// <returns>Result</returns>
        public WolframAlphaQueryResult GetWolframAlphaQueryResult(string response)
        {
            #region Conditions
            if (string.IsNullOrEmpty(response))
            {
                return new WolframAlphaQueryResult();
            }
            #endregion

            try
            {
                //load xmldocument from string
                var doc = new XmlDocument();
                doc.LoadXml(response);
                //find result section in xml document
                var mainNode = doc.SelectSingleNode("queryresult");
                //load result
                QueryResult = LoadResponse(doc);
                if (mainNode != null)
                {
                    //validate result
                    ValidationResult = new WolframAlphaValidationResult
                    {
#if MONOTOUCH || NET|| MONODROID
                        Success = mainNode.Attributes != null && bool.Parse(mainNode.Attributes["success"].Value),
                        Error = mainNode.Attributes != null && bool.Parse(mainNode.Attributes["error"].Value)
#endif
#if NETFX_CORE
                        Success = mainNode.Attributes != null && bool.Parse(mainNode.Attributes.GetNamedItem("success").NodeValue.ToString()),
                        Error = mainNode.Attributes != null && bool.Parse(mainNode.Attributes.GetNamedItem("error").NodeValue.ToString())
#endif
                    };
                    //get main part of answer with data
                    var selectSingleNode = mainNode.SelectSingleNode("assumptions");
                    if (selectSingleNode != null)
                    {
#if MONOTOUCH|| NET || MONODROID
                        var assumptions = selectSingleNode.ChildNodes;
                        foreach (XmlNode node in assumptions)
                        {
                            var assumption = new WolframAlphaAssumption
                                                 {
                                                     Word = node.Attributes["word"].Value,
                                                     Template = node.Attributes["template"].Value
                                                 };
                            foreach (XmlNode xmlNode in node.ChildNodes)
                            {
                                var desc = xmlNode.Attributes["desc"].Value;
                                var input = xmlNode.Attributes["input"].Value;
                                var name = xmlNode.Attributes[0].Value;
                                assumption.Categories.Add(new WolframAlphaAssumptionValue
                                {
                                    Desc = desc,
                                    Input = input,
                                    Name = name
                                });
                            }
#endif
#if NETFX_CORE
                        var assumptions = selectSingleNode.ChildNodes.Where(i => i.NodeName == "assumption").ToList();
                        foreach (var node in assumptions)
                        {
                            var assumption = new WolframAlphaAssumption
                                                 {
                                                     Word = node.Attributes[1].NodeValue.ToString(),
                                                     Template = node.Attributes[2].NodeValue.ToString()
                                                 };
                            foreach (var xmlNode in node.ChildNodes.Where(i => i.NodeName != "#text"))
                            {
                                var desc = xmlNode.Attributes.GetNamedItem("desc").NodeValue.ToString();
                                var input = xmlNode.Attributes.GetNamedItem("input").NodeValue.ToString();
                                var name = xmlNode.Attributes[0].NodeValue.ToString();
                                assumption.Categories.Add(new WolframAlphaAssumptionValue
                                                               {
                                                                   Desc = desc,
                                                                   Input = input,
                                                                   Name = name
                                                               });
                            }
#endif
                            QueryResult.WolframAlphaAssumptions.Add(assumption);
                        }
                    }
                }
            }
            #region Catch Excaptions
#if NET
            catch (XPathException ex)
            {
                WriteProblemResult(ex.Message);
            }
#endif
            catch (NullReferenceException ex)
            {
                WriteProblemResult(ex.Message);
            }
            catch (Exception ex)
            {
                WriteProblemResult(ex.Message);
            }
            #endregion

            if (Query != null && QueryResult != null)
            {
                QueryResult.Query = Query.Query;
                return QueryResult;
            }
            return null;
        }

        /// <summary>
        /// Get Query Result from XML
        /// </summary>
        /// <param name="response">Response strgin loaded in XmlDocument </param>
        /// <returns>Result</returns>
        public WolframAlphaQueryResult LoadResponse(XmlDocument response)
        {
            #region Conditions
            if (response == null)
            {
                return new WolframAlphaQueryResult();
            }
            #endregion
            try
            {
                var mainNode = response.SelectSingleNode("/queryresult");
                //validate result
                var queryResult = new WolframAlphaQueryResult
                {
                    Pods = new List<WolframAlphaPod>(),
#if MONOTOUCH|| NET|| MONODROID
                    Success = bool.Parse(mainNode.Attributes["success"].Value),
                    Error = bool.Parse(mainNode.Attributes["error"].Value),
                    NumberOfPods = Convert.ToInt32(mainNode.Attributes["numpods"].Value),
                    TimedOut = mainNode.Attributes["timedout"].Value,
                    DataTypes = mainNode.Attributes["datatypes"].Value
#endif
#if NETFX_CORE
                    Success = bool.Parse(mainNode.Attributes.GetNamedItem("success").NodeValue.ToString()),
                    Error = bool.Parse(mainNode.Attributes.GetNamedItem("error").NodeValue.ToString()),
                    NumberOfPods = Convert.ToInt32(mainNode.Attributes.GetNamedItem("numpods").NodeValue.ToString()),
                    TimedOut = mainNode.Attributes.GetNamedItem("timedout").NodeValue.ToString(),
                    DataTypes = mainNode.Attributes.GetNamedItem("datatypes").NodeValue.ToString()
#endif
                };
                //get pods and subpods
#if MONOTOUCH || NET|| MONODROID
                foreach (XmlNode node in mainNode.SelectNodes("pod"))
#endif
#if NETFX_CORE
                foreach (var node in mainNode.SelectNodes("pod"))
#endif
                {
                    var pod = new WolframAlphaPod
                    {
                        SubPods = new List<WolframAlphaSubPod>(),
#if MONOTOUCH || NET|| MONODROID
                        Title = node.Attributes["title"].Value,
                        Scanner = node.Attributes["scanner"].Value,
                        Position = Convert.ToInt32(node.Attributes["position"].Value),
                        Error = bool.Parse(node.Attributes["error"].Value),
                        NumberOfSubPods = Convert.ToInt32(node.Attributes["numsubpods"].Value),
#endif
#if NETFX_CORE
                        Title = node.Attributes.GetNamedItem("title").NodeValue.ToString(),
                        Scanner = node.Attributes.GetNamedItem("scanner").NodeValue.ToString(),
                        Position = Convert.ToInt32(node.Attributes.GetNamedItem("position").NodeValue.ToString()),
                        Error = bool.Parse(node.Attributes.GetNamedItem("error").NodeValue.ToString()),
                        NumberOfSubPods = Convert.ToInt32(node.Attributes.GetNamedItem("numsubpods").NodeValue),
#endif

                    };
#if  MONOTOUCH || NET || MONODROID
                    foreach (XmlNode subNode in node.SelectNodes("subpod"))
#endif
#if NETFX_CORE
                    foreach (var subNode in node.SelectNodes("subpod"))
#endif
                    {
                        var subPod = new WolframAlphaSubPod
                        {
#if  MONOTOUCH || NET || MONODROID
                            Title = subNode.Attributes["title"].Value
#endif
#if NETFX_CORE
                            Title = subNode.Attributes.GetNamedItem("title").NodeValue.ToString()
#endif
                        };
#if  MONOTOUCH || NET || MONODROID
                        foreach (XmlNode nodeContentNode in subNode.SelectNodes("plaintext"))
#endif
#if NETFX_CORE

                        foreach (var nodeContentNode in subNode.SelectNodes("plaintext"))
#endif
                        {
                            subPod.PodText = nodeContentNode.InnerText;
                        }
#if  MONOTOUCH || NET || MONODROID
                        foreach (XmlNode contentNode in subNode.SelectNodes("img"))
#endif
#if NETFX_CORE
                        foreach (var contentNode in subNode.SelectNodes("img"))
#endif
                        {
                            subPod.PodImage = new WolframAlphaImage
                            {
#if  MONOTOUCH || NET || MONODROID
                                Location = new Uri(contentNode.Attributes["src"].Value),
                                HoverText = contentNode.Attributes["alt"].Value,
                                Title = contentNode.Attributes["title"].Value,
                                Width = Convert.ToInt32(contentNode.Attributes["width"].Value),
                                Height = Convert.ToInt32(contentNode.Attributes["height"].Value)
#endif
#if NETFX_CORE
                                Location = new Uri(contentNode.Attributes.GetNamedItem("src").NodeValue.ToString()),
                                HoverText = contentNode.Attributes.GetNamedItem("alt").NodeValue.ToString(),
                                Title = contentNode.Attributes.GetNamedItem("title").NodeValue.ToString(),
                                Width = Convert.ToInt32(contentNode.Attributes.GetNamedItem("width").NodeValue.ToString()),
                                Height = Convert.ToInt32(contentNode.Attributes.GetNamedItem("height").NodeValue.ToString())
#endif
                            };
                        }
                        pod.SubPods.Add(subPod);
                    }
                    queryResult.Pods.Add(pod);
                    queryResult.SearchQuery = Query.Query;
                    QueryResult = queryResult;
                }
                return queryResult;
            }
            #region Catch Excaptions
#if  NET
            catch (XPathException ex)
            {
                WriteProblemResult(ex.Message);
                return new WolframAlphaQueryResult();
            }
#endif
            catch (NullReferenceException ex)
            {
                WriteProblemResult(ex.Message);
                return new WolframAlphaQueryResult();
            }
            catch (Exception ex)
            {
                WriteProblemResult(ex.Message);
                return new WolframAlphaQueryResult();
            }
            #endregion
        }
#endif

#if WINDOWS_PHONE
        public void BeginGetStringRequest(WolframAlphaQuery query)
        {
            if (query == null || string.IsNullOrEmpty(query.APIKey) || string.IsNullOrEmpty(query.Query))
            {
                return;
            }

            Query = query;
            //Get xml request from the server
            try
            {
                var webRequest = (HttpWebRequest) (WebRequest.Create(query.FullQueryString));
                webRequest.BeginGetRequestStream(ReadWebRequestCallback, webRequest);
            }
            catch(Exception ex)
            {
                Debug.WriteLine(ex.Message);
            }
        }

        private void ReadWebRequestCallback(IAsyncResult callbackResult)
        {
            try
            {
                HttpWebRequest myRequest = (HttpWebRequest) callbackResult.AsyncState;
                HttpWebResponse myResponse = (HttpWebResponse) myRequest.EndGetResponse(callbackResult);

                using (StreamReader httpwebStreamReader = new StreamReader(myResponse.GetResponseStream()))
                {
                    QueryResultString = httpwebStreamReader.ReadToEnd();
                }
                myResponse.Close();
            }
            catch(Exception ex)
            {
                Debug.WriteLine(ex.Message);
            }
        }

        public WolframAlphaQueryResult GetWolframAlphaQueryResult(string response)
        {
            #region Conditions
            if (string.IsNullOrEmpty(response))
            {
                return new WolframAlphaQueryResult();
            }
            #endregion

            try
            {
                //load xmldocument from string
                MemoryStream mStream = new MemoryStream(UnicodeEncoding.Unicode.GetBytes(response));
                var doc = XDocument.Load(mStream);

                //find result section in xml document
                var mainNode = doc.Descendants("queryresult").First();
                //load result
                QueryResult = LoadResponse(doc);
                if (mainNode != null)
                {
                    //validate result
                    ValidationResult = new WolframAlphaValidationResult
                                           {
                                               Success = mainNode.HasAttributes &&   bool.Parse(mainNode.Attribute("success").Value),
                                               Error =  mainNode.HasAttributes &&  bool.Parse(mainNode.Attribute("error").Value)
                                           };
                    //get main part of answer with data
                    var selectSingleNode = mainNode.Descendants("assumptions").First();
                    if (selectSingleNode != null)
                    {

                        var assumptions = selectSingleNode.Elements().ToList();
                        foreach (XElement node in assumptions)
                        {
                            var assumption = new WolframAlphaAssumption
                                                 {
                                                     Word = node.Attribute("word").Value,
                                                     Template = node.Attribute("template").Value
                                                 };
                            foreach (XElement xmlNode in node.Elements())
                            {
                                var desc = xmlNode.Attribute("desc").Value;
                                var input = xmlNode.Attribute("input").Value;
                                var name = xmlNode.FirstAttribute.Value;
                                assumption.Categories.Add(new WolframAlphaAssumptionValue
                                                              {
                                                                  Desc = desc,
                                                                  Input = input,
                                                                  Name = name
                                                              });
                            }

                            QueryResult.WolframAlphaAssumptions.Add(assumption);
                        }
                    }
                }
            }
            #region Catch Excaptions


            catch (NullReferenceException ex)
            {
                WriteProblemResult(ex.Message);
            }
            catch (Exception ex)
            {
                WriteProblemResult(ex.Message);
            }
            #endregion

            if (Query != null && QueryResult != null)
            {
                QueryResult.Query = Query.Query;
                return QueryResult;
            }
            return null;
        }

        public WolframAlphaQueryResult LoadResponse(XDocument response)
        {
            #region Conditions
            if (response == null)
            {
                return new WolframAlphaQueryResult();
            }
            #endregion
            try
            {
                var mainNode = response.Descendants("queryresult").First();
                //validate result
                var queryResult = new WolframAlphaQueryResult
                {
                    Pods = new List<WolframAlphaPod>(),
                    Success = bool.Parse(mainNode.Attribute("success").Value),
                    Error = bool.Parse(mainNode.Attribute("error").Value),
                    NumberOfPods = Convert.ToInt32(mainNode.Attribute("numpods").Value),
                    TimedOut = mainNode.Attribute("timedout").Value,
                    DataTypes = mainNode.Attribute("datatypes").Value

                };
                //get pods and subpods
                foreach (var node in mainNode.Descendants("pod").ToList())
                {
                    var pod = new WolframAlphaPod
                    {
                        SubPods = new List<WolframAlphaSubPod>(),
                        Title = node.Attribute("title").Value,
                        Scanner = node.Attribute("scanner").Value,
                        Position = Convert.ToInt32(node.Attribute("position").Value),
                        Error = bool.Parse(node.Attribute("error").Value),
                        NumberOfSubPods = Convert.ToInt32(node.Attribute("numsubpods").Value),
                    };
                    foreach (var subNode in node.Descendants("subpod"))
                    {
                        var subPod = new WolframAlphaSubPod
                        {
                            Title = subNode.Attribute("title").Value
                        };
                        foreach (var nodeContentNode in subNode.Descendants("plaintext").ToList())
                        {
                            subPod.PodText = nodeContentNode.Value;
                        }

                        foreach (var contentNode in subNode.Descendants("img").ToList())
                        {
                            subPod.PodImage = new WolframAlphaImage
                            {
                                Location = new Uri(contentNode.Attribute("src").Value),
                                HoverText = contentNode.Attribute("alt").Value,
                                Title = contentNode.Attribute("title").Value,
                                Width = Convert.ToInt32(contentNode.Attribute("width").Value),
                                Height = Convert.ToInt32(contentNode.Attribute("height").Value)
                            };
                        }
                        pod.SubPods.Add(subPod);
                    }
                    queryResult.Pods.Add(pod);
                    queryResult.SearchQuery = Query.Query;
                    QueryResult = queryResult;
                }
                return queryResult;
            }
            #region Catch Excaptions
            catch (NullReferenceException ex)
            {
                WriteProblemResult(ex.Message);
                return new WolframAlphaQueryResult();
            }
            catch (Exception ex)
            {
                WriteProblemResult(ex.Message);
                return new WolframAlphaQueryResult();
            }
            #endregion
        }
#endif
        #region Private Methods
        #region WriteProblemResult
        /// <summary>
        /// WriteProblemResult in Console
        /// </summary>
        /// <param name="message">Message to write</param>
        private static void WriteProblemResult(string message)
        {
#if NET
            Console.ForegroundColor = ConsoleColor.Red;
            Console.WriteLine(message);
            Console.ForegroundColor = ConsoleColor.Green;
#endif

#if Monodroid
            Console.Out.WriteLine(message);
#endif
        }
        #endregion
        #endregion
    }
}